"""
Defines the PrintConsumption class.
"""

import logging
import socket
import xml.etree.ElementTree as ET

from pika.spec import BasicProperties

from .execution_request import parse_xml
from .psquared_engine import Engine, RestfulFailure
from . import utils


class PSquaredConsumption:
    """
    This implementation of the rabbitmq-consume Consumption Protocol
        the consume PSquared messages and satifying the execution
        request contained within.
    """

    def __init__(
        self,
        _: BasicProperties,
        body: bytes,
        delivery_tag: int,
        redelivered: bool,
    ):
        """
        Creates a Consumption instance of the class that will consume
        the body.

        Args:
            properties: The properties of the message.
            body: the XML body of the RabbitMQ message to be consumed
                by the created Consumption instance.
            delivery_tag: The tag identifying the RabbitMQ message.
            redelivered: True if the supplied body has been delivered
                before.
        """
        self.__document = ET.fromstring(body)
        self.__delivery_tag = delivery_tag
        self.__redelivered = redelivered

    def consume(self) -> None:  #  pylint: disable=too-many-locals
        """
        Consumes the XML body of a RabbitMQ message.
        """
        logging.info("BEGIN CONSUMPTION")

        try:
            action = self.__document.get("action")
            if "ignore" == action:
                logging.info(
                    'Ignoring execution request as specified by its "action" element'
                )
            else:
                request = parse_xml(self.__document)
                logging.debug("Beginning state machine completion")
                engine = Engine()
                engine.execute(self.__execution_message(), request, self.__redelivered)
                logging.info("END CONSUMPTION")
        except RestfulFailure as restful_failure:
            logging.error("Failed to complete state machine")
            logging.exception(restful_failure)
            utils.eprint("Failed in PSquared:\n", restful_failure)
            logging.info("FAIL CONSUMPTION")
        except Exception as abort_exception:  #  pylint: disable=broad-except
            logging.error("Aborted while in state machine")
            logging.exception(abort_exception)
            utils.eprint("Aborted in PSquared:\n", abort_exception)
            logging.info("ABORT CONSUMPTION")

    def delivery_tag(self) -> int:
        """
        Returns the tag identifying the RabbitMQ message.
        """
        return self.__delivery_tag

    def __execution_message(self) -> str:
        return "Executing on " + socket.gethostname()



"""Provide utilities to help using the psquared_consume package"""

from typing import Dict

from argparse import Namespace
import os
import sys


def eprint(*args, **kwargs) -> None:
    """
    Print message to stderr
    """
    print(*args, file=sys.stderr, **kwargs)


def pretty_print(uri: str, document: str, response=True):
    """
    Simple pretty print for *IX systems to the error stream.

    Args:
        uri: The uri to which the request was made.
        document: The XML document to print.
        response: True if the document is a response to a request,
                otherwise it is treated as the payload to a request.
    """
    if None is not uri:
        if response:
            eprint("URI : Response : " + uri)
        else:
            eprint("URI : Request :  " + uri)
    #    print(document)
    os.system(f"echo '{document}' | xmllint -format - 1>&2")
    eprint("--------")


def read_envar_values(mapping: Dict[str, str]):
    """
    Create a argparse.Namespace instance populated by the values of the
    envrionmental variables specified by the keys of the mapping.
    """
    result = {}
    for key in mapping.keys():
        value = os.getenv(key)
        if None is not value:
            option = mapping[key]
            result[option] = value
    return Namespace(**result)



"""
Provide utilities to run the PSqaured state machine
"""

from typing import List, Tuple, Optional

import io
import os
import subprocess
import tempfile
import xml.etree.ElementTree as ET

import requests

from .execution_request import ExecutionRequest
from . import mock_server
from . import utils

HEADERS = {"Content-Type": "application/xml", "Accept": "application/xml"}
DEBUG = False
RESUBMISSION_LIMIT = 8

EXAMPLE_HTTP = "http://example.com/"
EXAMPLE_HTTPS = "https://example.com/"
EXAMPLE_URIS = (EXAMPLE_HTTP, EXAMPLE_HTTPS)


class RestfulFailure(Exception):
    """
    Thrown when a RESTful call does not complete as expected.
    """

    def __init__(self, message: str, error_code: int):
        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, message)
        self.code = error_code


def _check_status(uri: str, response: requests.Response, expected: int) -> None:
    """
    Checks that the returned status code matches the expected one.

    Args:
        uri: The URI to which the request was made.
        response: the response to the request.
        expected: the expected status code.
    """
    if 400 == response.status_code:
        raise RestfulFailure(
            'Application at "' + uri + '" requires SSL certificate',
            response.status_code,
        )
    if 401 == response.status_code:
        raise RestfulFailure(
            'Not authorized to execute commands for Application at "' + uri,
            response.status_code,
        )
    if 404 == response.status_code:
        raise RestfulFailure(
            'Application at "' + uri + '" not found', response.status_code
        )
    if expected != response.status_code:
        raise RestfulFailure(
            f'Unexpected status ({str(response.status_code)}) returned from "{uri}"',
            response.status_code,
        )


def _get_valid_element(uri: str, element: ET.Element, tag: str) -> ET.Element:
    """
    Returns the valid tag element within the supplied element.

    Raises:
        RestfulFailure: When the supplied element does not contain
                the tag element.
    """
    tag_element = element.find(tag)
    if None is tag_element:
        raise RestfulFailure(
            f'Incomplete response returned from "{uri}", "{tag}" element is missing from'
            + ' "{element.tag}" element',
            1,
        )
    return tag_element


def _get_valid_text(uri: str, element: ET.Element, tag: str) -> str:
    """
    Returns the valid, non-null, contents for the tag element within the supplied element.

    Raises:
        RestfulFailure: When the supplied element does not contain
                the tag element or the tag element is null.
    """
    tag_element = element.find(tag)
    if None is tag_element or None is tag_element.text:
        raise RestfulFailure(
            f'Incomplete response returned from "{uri}", "{tag}" element is missing from'
            + ' "{element.tag}" element, or is empty',
            1,
        )
    return tag_element.text


def _prepare_attachment(message: Optional[str] = None) -> ET.Element:
    """
    Returns the 'attachment' document that includes the supplied message, if any.
    """
    attachment = ET.Element("attachment")
    if None is not message:
        msg = ET.Element("message")
        msg.text = message
        attachment.append(msg)
    return attachment


def _process(
    cmd: str,
    params: List[str],
    cmd_out: Optional[io.TextIOWrapper],
    cmd_err: Optional[io.TextIOWrapper],
) -> Tuple[int, str]:
    """
    Processes the supplied command, return any message created by the command while
    it was executing.

    Args:
        cmd: The command to process.
        params: The parameter to use with the command.
        cmd_out: Where to write the stdout of the command.
        cmd_out: Where to write the stderr of the command.
    """

    args = []
    args.append(cmd)
    args.extend(params)
    file_descriptor, message_file = tempfile.mkstemp(text=True)
    cmd_env = os.environ.copy()
    cmd_env["PP_MESSAGE_FILE"] = message_file
    with subprocess.Popen(args, env=cmd_env, stdout=cmd_out, stderr=cmd_err) as proc:
        result = proc.wait()
    message = os.fdopen(file_descriptor, "r").read()
    os.remove(message_file)
    return result, message.strip()


class Engine:  # pylint: disable=too-few-public-methods
    """
    This class guides a ExecutionRequest though the PSquared state machine.
    """

    def __init__(
        self,
        debug: Optional[bool] = None,
        cert: Optional[str] = None,
        key: Optional[str] = None,
        cacert: Optional[str] = None,
    ):
        """
        Creates an instance of this class.

        Args:
            cert: The path to the file holding this client's
                certificate.
            key: The path to the file holding this client's private
                key.
            cacert: The path to the known CA certificates.
        """
        self.__debug = debug
        self.__session = requests.Session()
        if (
            None is not cert
            and os.path.exists(cert)
            and None is not key
            and os.path.exists(key)
        ):
            self.__session.cert = (cert, key)
        if None is not cacert and os.path.exists(cacert):
            self.__session.verify = cacert
        self.__use_mock_server = False

    def __automatic_resubmission(self, uri: str, resubmission_count: int) -> str:
        """
        Resubmits this request as its previous engine failed top complete.

        Raise:
            ConnectionFailure: When the connection to the server can not be established.
        """
        if self.__use_mock_server:
            response = mock_server.get(uri)
        else:
            response = self.__session.get(uri)
        _check_status(uri, response, 200)
        if self.__debug:
            utils.pretty_print(uri, response.text)
        state = ET.fromstring(response.text)
        elements = state.findall("exits/exit")
        if 0 == len(elements):
            element = _get_valid_element(uri, state, "exited")
            name_text = _get_valid_text(uri, element, "name")
            if "reschedule" == name_text:
                uri_text = _get_valid_text(uri, element, "uri")
                return self.__get_execution_uri(uri_text, resubmission_count)
            raise RestfulFailure(
                "Another process has started processing this request, so this attempt will halt",
                409,
            )
        for exit_element in elements:
            name_text = _get_valid_text(uri, exit_element, "name")
            if "reschedule" == name_text:
                uri_text = _get_valid_text(uri, element, "uri")
                reschedule_uri = uri_text
                break

        state = self.__execute_transition(
            reschedule_uri, "Rescheduling as previous execution died"
        )
        elements = state.findall("exits/exit")
        if 0 == len(elements):
            raise RestfulFailure("Failed to correctly reschedule submission", 1)
        return self.__parse_submitted_response(uri, state, resubmission_count)

    def execute(  # pylint: disable=too-many-branches
        self, message_for_execute: str, request: ExecutionRequest, resubmission: bool
    ) -> None:
        """
        Executes an instance of the PSquared state machine.

        Args:
            request: the execution request to fulfill.
            message_for_execute: The message to associate with the
                "execute" transition.
        Raises:

        """

        if resubmission:
            resubmission_count = 0
        else:
            resubmission_count = -1
        executing_uri = self.__get_execution_uri(
            request.submission_uri, resubmission_count
        )
        processed_uri, failed_uri = self.__processing_beginning(
            executing_uri, message_for_execute
        )
        if None is request.arguments:
            params: List[str] = []
        else:
            params = request.arguments
        try:
            result, message = _process(
                request.cmd, params, request.cmd_out, request.cmd_err
            )
        except OSError as os_error:
            utils.eprint(request.cmd, params)
            utils.eprint("Failed to execute command")
            if None is not os_error.errno:
                utils.eprint("Error Number: " + str(os_error.errno))
            if None is not os_error.strerror:
                utils.eprint("Explanation: " + os_error.strerror)
            if None is not os_error.filename:
                utils.eprint("Filename: " + os_error.filename)
            result, message = (
                255,
                "Failed to execute command, check log file for details",
            )
        if 0 == result:
            self.__execute_transition(processed_uri, message)
            if None is not request.success_cmd:
                try:
                    _process(
                        request.success_cmd, params, request.cmd_out, request.cmd_err
                    )
                except OSError as exc:
                    raise RestfulFailure(
                        'Failed to execute "success" command', 255
                    ) from exc
        else:
            self.__execute_transition(failed_uri, message)
            if None is not request.failure_cmd:
                try:
                    _process(
                        request.failure_cmd, params, request.cmd_out, request.cmd_err
                    )
                except OSError as exc:
                    raise RestfulFailure(
                        'Failed to execute "failure" command', 255
                    ) from exc

    def __execute_transition(
        self, uri: str, message: Optional[str] = None
    ) -> ET.Element:
        """
        Requests the PSquared instance create a new realized state.

        Args:
            uri: The URI of the destination of the transition.
            message: The message, if any, to attach to the transition.

        Returns:
            The XML document that was the response to the request.

        Raise:
            RestfulFailure: If the status code of the response is not what was expected.
        """
        # Prepare attachment document
        attachment = _prepare_attachment(message)
        if self.__debug:
            utils.pretty_print(uri, str(ET.tostring(attachment), "utf-8"), False)
        if self.__use_mock_server:
            response = mock_server.post(uri)
        else:
            response = self.__session.post(
                uri, data=ET.tostring(attachment), headers=HEADERS
            )
        if 404 == response.status_code:
            raise RestfulFailure(
                f'Exit request "{uri}" not found', response.status_code
            )
        if 409 == response.status_code:
            raise RestfulFailure(
                "Another process has changed the processing of this request, so this attempt will"
                + " halt",
                response.status_code,
            )
        if 201 == response.status_code:
            if self.__debug:
                utils.pretty_print(uri, response.text)
            return ET.fromstring(response.text)
        raise RestfulFailure(
            f'Unexpected status ({str(response.status_code)}) returned from "{uri}"',
            response.status_code,
        )

    def __get_execution_uri(self, uri: str, resubmission_count: int) -> str:
        """
        Returns the URI that will signal to PSqaured that processing has begun.

        Args:
            uri: The URL associated with the submission of execution request.

        Raises:
            ConnectionFailure: When the connection to the server can not be established.
        """
        if uri.startswith(EXAMPLE_URIS):
            self.__use_mock_server = True
            response = mock_server.get(uri)
        else:
            response = self.__session.get(uri)
        _check_status(uri, response, 200)
        if self.__debug:
            utils.pretty_print(uri, response.text)
        state = ET.fromstring(response.text)
        return self.__parse_submitted_response(uri, state, resubmission_count)

    def __parse_submitted_response(
        self, uri: str, state: ET.Element, resubmission_count: int
    ) -> str:
        """
        Raises:
                RestfulFailure: If the status code of the response is not what was expected.
        """
        elements = state.findall("exits/exit")
        if 0 == len(elements):
            element = _get_valid_element(uri, state, "exited")
            if not 0 > resubmission_count:
                name_text = _get_valid_text(uri, element, "name")
                if "executing" == name_text:
                    if resubmission_count == RESUBMISSION_LIMIT:
                        raise RestfulFailure(
                            "re-submission limit has been reached, so this attempt will halt",
                            400,
                        )
                    uri_text = _get_valid_text(uri, element, "uri")
                    return self.__automatic_resubmission(
                        uri_text, resubmission_count + 1
                    )
                raise RestfulFailure(
                    "State that follows submission does not allow automatic re-submission, so this"
                    + " attempt will halt",
                    409,
                )
            raise RestfulFailure(
                "Another process has started processing this request, so this attempt will halt",
                409,
            )
        for exit_element in elements:
            name_text = _get_valid_text(uri, exit_element, "name")
            if "executing" == name_text:
                return _get_valid_text(uri, exit_element, "uri")
        raise RestfulFailure(
            f'Incomplete response returned from "{uri}", no "exit" element named "executing"',
            1,
        )

    def __processing_beginning(self, uri: str, message: str) -> Tuple[str, str]:
        """
        Informs the PSquared instance that processing is beginning for this
        item/version pairing.
        """
        state = self.__execute_transition(uri, message)
        elements = state.findall("exits/exit")
        for exit_element in elements:
            name_text = _get_valid_text(uri, exit_element, "name")
            if "processed" == name_text:
                processed_uri = _get_valid_text(uri, exit_element, "uri")
                break

        for exit_element in elements:
            name_text = _get_valid_text(uri, exit_element, "name")
            if "failed" == name_text:
                failed_uri = _get_valid_text(uri, exit_element, "uri")
                break

        return processed_uri, failed_uri



"""Provides mock responses for exercising the Engine"""

from typing import ByteString, Optional

import requests


def get(uri: str) -> requests.Response:
    """
    Return a mock response to a GET request, based on the supplied uri.
    """
    result = requests.Response()
    match uri:
        case _:
            text = """<realized-state>
    <exits>
        <exit>
            <name>executing</name>
            <uri>https://example.com/psquared/local/command/exit/1/executing</uri>
        </exit>
    </exits>
</realized-state>"""
    result.raw = MockResponse(text)
    result.status_code = 200
    return result


def post(uri: str) -> requests.Response:
    """
    Return a mock response to a POST request, based on the supplied uri.
    """
    result = requests.Response()
    if uri.endswith("executing"):
        text = """<realized-state>
    <exits>
        <exit>
            <name>processed</name>
            <uri>https://example.com/psquared/local/command/exit/2/processed</uri>
        </exit>
        <exit>
            <name>failed</name>
            <uri>https://example.com/psquared/local/command/exit/2/failed</uri>
        </exit>
    </exits>
</realized-state>"""
    elif uri.endswith("processed"):
        text = """<realized-state>
    <exits>
        <exit>
            <name>reset</name>
            <uri>https://example.com/psquared/local/command/exit/2/reset</uri>
        </exit>
    </exits>
</realized-state>"""

    else:
        raise ValueError("unknown URI")
    result.raw = MockResponse(text)
    result.status_code = 201
    return result


class MockResponse:  # pylint: disable=too-few-public-methods
    """
    This is used to mock up an HTTP response when testing this package.
    """

    def __init__(self, message: str):
        """
        Creates an instance of this class.

        Args:
            message: The string to return as the contents of this object.
        """
        self.__content = message.encode()
        self.__index = 0

    def read(self, amt: int) -> Optional[ByteString]:
        """
        Read the request amount of data
        """
        length = len(self.__content)
        if self.__index >= length:
            return None
        begin = self.__index
        end = self.__index + amt
        end = min(end, length)
        self.__index = end - begin
        return self.__content[begin:end]



"""A command line interface to run an instance of the PSquared state machine."""

import argparse
import logging
import socket
import sys

import requests

from .execution_request import ExecutionRequest
from .psquared_engine import Engine, RestfulFailure
from . import utils

_ENVARS_MAPPING = {
    "PP_ENGINE_LOG_FILE": "LOG_FILE",
    "PP_ENGINE_LOG_LEVEL": "LOG_LEVEL",
}

_LOG_LEVELS = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "NOTSET": logging.NOTSET,
}


def _create_argument_parser() -> argparse.ArgumentParser:
    """
    Creates and populated the argparse.ArgumentParser for this executable.
    """
    parser = argparse.ArgumentParser(
        description="Executes an instance of a rabbit_consume.pusher class"
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="DEBUG",
        help="print out detail information to stdout, automatically set log level to DEBUG.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--log_file",
        dest="LOG_FILE",
        help="The file, as opposed to stdout, into which to write log messages",
    )
    parser.add_argument(
        "-l",
        "--log_level",
        default="INFO",
        dest="LOG_LEVEL",
        help="The logging level for this execution",
        choices=_LOG_LEVELS.keys(),
    )
    parser.add_argument(
        "-s",
        "--success_cmd",
        dest="SUCCESS_CMD",
        help="The command to run after a successful processing",
        default=None,
    )
    parser.add_argument(
        "-f",
        "--failure_cmd",
        dest="FAILURE_CMD",
        help="The command to run after a failed processing",
        default=None,
    )
    parser.add_argument(
        "-o",
        dest="OUTPUT",
        type=argparse.FileType("a"),
        help="The file, as opposed to stdout, into which to write the output of the command",
        default=None,
    )
    parser.add_argument(
        "-e",
        dest="ERROR",
        type=argparse.FileType("a"),
        help="The file, as opposed to stderr, into which to write errors from the command",
        default=None,
    )
    parser.add_argument(
        "-r",
        "--resubmission",
        dest="RESUBMISSION",
        help="allows automatic re-submission to take place.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "submission_url",
        help="The submission URL used to create the processing request to be handled",
    )
    parser.add_argument("cmd", help="The command to be executed")
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        help="An argument to the command to be executed",
    )
    return parser


def main():
    """
    A command line interface to run an instance of the PSquared state machine.
    """
    parser = _create_argument_parser()
    envar_values = utils.read_envar_values(_ENVARS_MAPPING)
    options = parser.parse_args(namespace=envar_values)

    if None is options.LOG_FILE:
        logging.basicConfig(stream=sys.stdout, level=_LOG_LEVELS[options.LOG_LEVEL])
    else:
        logging.basicConfig(
            filename=options.LOG_FILE, level=_LOG_LEVELS[options.LOG_LEVEL]
        )

    logging.debug("Begin options:")
    for option in options.__dict__:
        if options.__dict__[option] is not None:
            logging.debug("    %s = %s", option, options.__dict__[option])
    logging.debug("End options:")

    logging.info("BEGIN CONSUMPTION")
    try:
        request = ExecutionRequest(
            options.submission_url,
            options.cmd,
            options.args,
            options.SUCCESS_CMD,
            options.FAILURE_CMD,
            options.OUTPUT,
            options.ERROR,
        )
        engine = Engine(options.DEBUG)
        engine.execute(
            "Executing on " + socket.gethostname() + " using pp_engine.py",
            request,
            options.RESUBMISSION,
        )
        logging.info("END CONSUMPTION")
    except requests.exceptions.ConnectionError:
        logging.error(
            "Could not connection to PSquared server with url %s",
            options.submission_url,
        )
        logging.info("ABORT CONSUMPTION")
    except RestfulFailure as psquared_failure:
        logging.error("Failed to complete state machine")
        logging.exception(psquared_failure)
        utils.eprint("Failed in PSquared:\n", psquared_failure)
        logging.info("FAIL CONSUMPTION")
    except Exception as abort_exception:  #  pylint: disable=broad-except
        logging.error("Aborted while in state machine")
        logging.exception(abort_exception)
        utils.eprint("Aborted in PSquared:\n", abort_exception)
        logging.info("ABORT CONSUMPTION")



"""
Defined the ExecutionRequest class.
"""

from typing import List, Optional

import errno
import io
import os
import xml.etree.ElementTree as ET


class RestfulFailure(Exception):
    """
    Thrown when document can not be parses to create a
        ExecutionRequest instance.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


def _build_string(element: Optional[ET.Element]) -> Optional[str]:
    """
    Builds a string from a ElementTree Element.

    Args:
        element: the ElementTree Elements from which to build the
                result.
    """
    if None is element or "" == element.text:
        return None
    return element.text


def _build_strings(
    elements: Optional[List[ET.Element]],
) -> Optional[List[str]]:
    """
    Builds a List of string from a set of ElementTree Elements.

    Args:
        args: the List of ElementTree Elements from which to build
                the result.
    Returns:
        the List of string from a set of ElementTree Elements.

    Raises:
        ValueError: If any of the string are None.
    """
    if None is elements or 0 == len(elements):
        return None
    arguments: List[str] = []
    for element in elements:
        text = _build_string(element)
        if None is text:
            raise ValueError("The string are not optional in the List")
        arguments.append(text)
    return arguments


def _mkdirs(path: str) -> None:
    """
    Makes sure the supplied directory exists, creating it if necessary.

    Args:
        path: The path to the directory.
    """
    try:
        os.makedirs(path)
    except OSError as os_error:
        if os_error.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise os_error


def _open_file(path: str) -> Optional[io.TextIOWrapper]:
    """
    Opens a file at the supplied path, creating any directory that does
    not exist. Relative paths are create with respect to the HOME
    directory of the user.

    Args:
        path: The path to the file.
    """
    if path.startswith("/"):
        filepath = path
    else:
        filepath = os.path.expanduser("~/" + path)
    directory = os.path.dirname(filepath)
    if None is not directory and "" != directory:
        _mkdirs(directory)
    return open(filepath, "w")  # pylint: disable=unspecified-encoding


class ExecutionRequest:  # pylint: disable=too-few-public-methods
    """
    This class encapsulates a requests from PSquared to execute a command.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        submission_uri: str,
        cmd: str,
        arguments: Optional[List[str]] = None,
        success_cmd: Optional[str] = None,
        failure_cmd: Optional[str] = None,
        cmd_out: Optional[io.TextIOWrapper] = None,
        cmd_err: Optional[io.TextIOWrapper] = None,
    ):
        """
        Create an instance of this class.

        Args:
            submission_url: The URI of the submission request for
                this execution.
            cmd: The command to be executed in order to satisfy this
                request.
            arguments: The sequence of arguments to be passed to the
                commands.
            success_cmd: The command, if any, to execute if this
                request succeeds.
            failure_cmd: The command, if any, to execute if this
                request fails.
            cmd_out: The path, if any, to use at the standard output
                when executing the request.
            cmd_err: The path, if any, to use at the standard error
                when executing the request.
        """
        self.submission_uri = submission_uri
        self.cmd = cmd
        self.success_cmd = success_cmd
        self.failure_cmd = failure_cmd
        self.arguments = arguments
        self.cmd_out = cmd_out
        self.cmd_err = cmd_err


def parse_xml(document: ET.Element) -> ExecutionRequest:
    """
    Creates an ExecutionRequest instance by parsing the supplied XML.

    Raises:
        ValueError: If the element does not contain a valid
                ExecutionRequest.
    """
    element = document.find("submission_uri")
    if None is element:
        raise ValueError('There is no "submission_uri" in the document')
    text = element.text
    if None is text or "" == text:
        raise ValueError('There is no "submission_uri" in the document')
    submission_uri: str = text

    element = document.find("command")
    if None is element:
        raise ValueError('There is no "command" in the document')
    text = element.text
    if None is text or "" == text:
        raise ValueError('There is no "command" in the document')
    cmd: str = text
    args = document.findall("argument")
    arguments = _build_strings(args)
    element = document.find("success_cmd")
    success_cmd = _build_string(element)
    element = document.find("failure_cmd")
    failure_cmd = _build_string(element)
    attribute = document.find("stdout")
    filepath = _build_string(attribute)
    if None is filepath:
        cmd_out = None
    else:
        cmd_out = _open_file(filepath)
    attribute = document.find("stderr")
    filepath = _build_string(attribute)
    if None is filepath:
        cmd_err = None
    else:
        cmd_err = _open_file(filepath)
    return ExecutionRequest(
        submission_uri,
        cmd,
        arguments,
        success_cmd,
        failure_cmd,
        cmd_out,
        cmd_err,
    )
